import levels
import math
import itertools
import collections
import functools
import time
import inspect
if __name__ == '__main__':
    print("This file is not meant to be executed. Please try main.py.")
    exit()
    
import pygame
#from pygame.locals import *
import pygcurse
import random

def null_func(*attrs):
    "A function that does absolutely nothing. Used as the default think for drones."
    pass

def movement_check(newpos,level):
    "Check if there is a collidable object at newpos. Returns None if it's okay to move; the blocking object if not."
    x,y = newpos
    hit=level.get_at((int(x),int(y)),True)

    if hit and not (isinstance(hit, no_collide_objs)):
        if isinstance(hit,Gate) and hit.open:
            return None
        else:
            return hit
        
def distance(a, b):
    "Return the distance between A and B's int_pos() if theyre gameobjs, them as 2-iters if not. If A or B is None, distance is returned as infinite."
    if not (a and b):
        return float("+inf")
    
    if isinstance(a, GameObj):
        ax, ay = a.int_pos()
    else:
        ax, ay = a
    
    if isinstance(b, GameObj):
        bx, by = b.int_pos()
    else:
        bx, by = b
        
    return math.sqrt((ax-bx)**2+(ay-by)**2)



def raycast(pos, vec, level):
    """Find the nearest object to the given position along the given vector tuple and return it. Works in the pygcurse 
    coordinate space (it'll look at cells, not pixels, unlike laser()). Try to avoid raycasting in a direction where there
    isn't anything. The direction vector must be normalized; otherwise bad things will happen. You may want to use Pygame's
    built-in vector class."""
    dx, dy = vec
    t=0
    x, y = pos
    obj=None
    while t < 5000:
        t+=1
        
        #@TODO: This'll check most tiles twice. Might be worth tweaking it so that
        #it doesn't do that.
        x+=dx
        #gamestate.mainwin.putchar(" ", round(x),round(y), bgcolor="blue")
        obj=level.get_at((round(x),round(y)),True)
        if obj: break
        
        y+=dy
        #gamestate.mainwin.putchar(" ", round(x),round(y), bgcolor="green")
        obj=level.get_at((round(x),round(y)),True)
        if obj: break 

    return obj

def raycast_to(src, dest, level):
    "Run a raycast from the src pos to the dest pos and return whatever the raycast hits"
    direction=pygame.math.Vector2(dest)-pygame.math.Vector2(src)
    direction.normalize_ip()
    return raycast(src, direction, level)

def laser(src, dest, gamestate):
    """Fire a deadly laser at tgt from src, where both are GameObj-s. This does a pixel-by-pixel raycast in the pygame vector space,
    not pygcurse's. It "breaks" the game model, sort of. Returns the GameObj the laser hit."""
    surface=gamestate.mainwin.surface
    src=gamestate.mainwin.gettopleftpixel(src)
    dest=gamestate.mainwin.gettopleftpixel(dest)
    srcx, srcy = src
    tgtx, tgty = dest
    
    #Make a direction vector from src to tgt and normalize it (make each element < 1)
    #Lasers are supposed to have a certain amount of inaccuracy, so we add a little random factor to the vector
    vec=pygame.math.Vector2(tgtx-srcx, tgty-srcy)
    vec.normalize_ip()
    randvec=pygame.math.Vector2(random.random()/10,random.random()/10)
    vec=vec+randvec
    
    x = srcx
    y = srcy
    
    dx, dy = vec

    black=pygame.Color(0,0,0)
    
    surface.lock()
    try:
        while True:
            x+=dx
            y+=dy
            px=surface.get_at((round(x),round(y)))
            pygame.display.flip()
            
            if px != black:
                break
    except IndexError:
        #If the trace extends off the screen, we'll get IndexError.
        pass
    finally:
        surface.unlock()
    
    
    pygpos=gamestate.mainwin.getcoordinatesatpixel(x,y)
    hit=gamestate.current_level.get_at(pygpos)
    color=pygame.Color(random.randint(0,255),random.randint(0,255),random.randint(0,255))
    gamestate.current_level.objlist.append(LaserBeam(src, (x,y), color))
    return hit
    
class GameObj:
    "The subclass from which all in-game objects are derived."
    
    def __init__(self,graphic=' ',pos=(-1,-1)):
        self.graphic=graphic #The char that is drawn to represent this obj
        self.pos=pos
    
    def update(self,gamestate):
        "Called each frame. Contains your graphics code."
        pass
    
    def interact(self,player,gamestate):
        """Called when the player tries to interact with this object (i.e. walks into the same square as it).
        This gets called as part of the player object's movement process, from its update(), so don't worry
        about race conditions or anything."""
        pass
    
    def trigger(self):
        """Called by a Trigger subclass instance. A touchplate, for example, would call trigger() on its target when
        the player steps on it. A gate, for example, will open when triggered."""
        pass
    
    def draw(self,window):
        "Called each frame to draw the object's character representation to the screen."
        x, y=self.pos
        #There are certain circumstances where an object will have non-integer pos, such as the boulder
        #having fraction velocity, and I'm not sure how pygcurse deals with that, so we just int-cast
        window.putchar(self.graphic, int(x), int(y))
    
    def int_pos(self):
        "Return this obj's pos with x and y cast to ints. Some objects can have float positions."
        return tuple([int(t) for t in self.pos])
    
class Player(GameObj):
    "The player object, controlled by the player. No shit, huh?"
    
    def __init__(self,startingpos):
        super().__init__('@',startingpos)
        #The level gives us our console when it starts up.
        self.injectionconsole=None
    
    def update(self,gamestate):
        #First figure out where we should move,
        #by grabbing Pygame keydown events.
        move_x=move_y=0
        for kbd in pygame.event.get(pygame.locals.KEYDOWN):
            key=kbd.key
            #yay elif chain yay
            if key == pygame.locals.K_UP:
                move_y=-1
            elif key == pygame.locals.K_DOWN:
                move_y=1
            elif key == pygame.locals.K_LEFT:
                move_x=-1
            elif key == pygame.locals.K_RIGHT:
                move_x=1
            #elif key == pygame.locals.K_SEMICOLON:
            elif kbd.unicode == ";":
                #Semicolon opens a pseudo interactive interpreter prompt that is used to... win the game
                self.injectionconsole.interact()
            elif key == pygame.locals.K_ESCAPE:
                #Yes, the player object handles making the game quit. No, it probably shouldn't. No, I don't care.
                gamestate.save_and_quit()
            elif key == pygame.locals.K_r:
                #R brings up the level reset dialogue
                gamestate.reset_level()
                return
        #Check that the controls aren't locked (this is unused, actually)
        if gamestate.controls_locked:
            return
        #We also don't want to do all this stuff if we haven't actually been asked to move
        #('interaction' only happens when the plyobj first walks onto the square)
        if not (move_x or move_y):
            return
        #We keep our last position around to make it easer for walls and such to stop us from walking into them.
        self.oldpos=self.pos
        x, y = self.pos
        self.pos=(x+move_x,y+move_y)
        #Now that we've moved, figure out who, if anyone, we've just walked into
        interactee=gamestate.current_level.get_at(self.pos)
        #If it's None, we moved into empty space. Otherwise, we tell them we just interacted with them.
        if interactee:
            interactee.interact(self,gamestate)
            
    def kill(self, src, gamestate):
        "'Kill' the player character (teleport him to his starting position)."
        #Death is cheap; all you really lose is physical progress through the level.
        self.pos=gamestate.current_level.plystartpos
        
class Wall(GameObj):
    
    def __init__(self,startingpos):
        super().__init__('#',startingpos)
        
    def interact(self,player,gamestate):
        player.pos=player.oldpos
        #print("Bumped into wall at",self.pos)
        
class Trigger(GameObj):
    
    def __init__(self,graphic,pos,target):
        super().__init__(graphic,pos)
        self.target=target
    
    def _trip(self):
        "Usually called by a Trigger's interact(). Causes the trigger's target to do whatever it does."
        self.target.trigger()
        
class Gate(GameObj):
    "An object that, when triggered, toggles whether or not the player can walk through it. He can't by default."
    
    def __init__(self,startingpos):
        super().__init__('█',startingpos)
        self.open=False
        
    def interact(self,player,gamestate):
        if not self.open:
            player.pos=player.oldpos
            
    def update(self,gamestate):
        if self.open:
            self.graphic=' '
        else:
            self.graphic='█'
            #We also make sure we can "close on" the player; if he's in the doorway when we close, he'll get bounced back
            if gamestate.current_level.plyobj.pos == self.pos:
                self.interact(gamestate.current_level.plyobj, gamestate)
            
    def trigger(self):
        self.open=not self.open
        
class Levelporter(GameObj):
    
    def __init__(self,startingpos,targetlevel):
        super().__init__('⌂',startingpos)
        self._targetlevel=targetlevel
        
    def interact(self,plyobj,gamestate):
        #Double check that the player is actually standing on us.
        if gamestate.current_level.plyobj.pos != self.pos:
            print("Nice try, but no.")
            return
        else:
            gamestate.change_level(self._targetlevel)
            
class PressurePlate(Trigger):
    "A trigger that fires whenever something enters or leaves its tile."
    
    def __init__(self,startingpos,target):
        super().__init__('⌧',startingpos,target)
        self.down=False
        
    def update(self,gamestate):
        if self.down:
            self.down=False
            wasdown=True
        else:
            wasdown=False
            
        for theobj in gamestate.current_level.objlist + gamestate.current_level.safe_objlist + [gamestate.current_level.plyobj]:
            #We use a list comprehension to int-cast theobj's x and y, in case it has float pos
            if theobj.int_pos() == self.pos and not theobj is self:
                self.down=True
        
        #If we were down but aren't, or are down but weren't, trip. If things stay the same, don't.        
        if not (self.down == wasdown):
            self._trip()
            
    def draw(self,win):
        #We don't want to draw ourself if we're down; player should see the thing on top of us instead.
        if not self.down:
            super().draw(win)
            
class Boulder(GameObj):
    "An object that moves according to its own idea of gravity and can trip pressureplates. It stops when it encounters most objects."
    
    def __init__(self,startingpos):
        super().__init__('○',startingpos)
        self.gravity=(0,1)
        
    def interact(self,plyobj,gamestate):
        plyobj.pos=plyobj.oldpos
        
    def update(self, gamestate):
        x, y = self.pos
        gravx, gravy = self.gravity
        
        #Things get hairy if we allow abs(gravity) > 1, so... we don't.
        if abs(gravx) > 1:
            gravx=gravx//abs(gravx)
        if abs(gravy) > 1:
            gravy=gravy//abs(gravy)
        x += gravx #Note that we don't actually set our position until after the collision check
        y += gravy
        
        #Movement_check returns None if we're not blocked.
        if movement_check((x,y),gamestate.current_level):
            return
        
        self.pos=(x,y)
        
class GravPad(GameObj):
    "An object that finds objects range tiles above it (the -y direction) that have a gravity attribute and sets that attribute to gravadj"
    
    def __init__(self, pos, gravadj, range):
        super().__init__("♨", pos)
        self.gravadj=gravadj
        self.range=range
        
    def update(self,gamestate):
        x, y = self.pos
        #If gravadj is invalid (not a tuple of two ints), we can't function.
        if not (isinstance(self.gravadj,tuple) and len(self.gravadj)==2 and 
                isinstance(self.gravadj[0],int) and isinstance(self.gravadj[1],int)):
            return
        #For every tile up to range tiles above us...
        for yofst in range(1,self.range+1):
            #See what's there...
            theobj=gamestate.current_level.get_at((x, y-yofst))
            #If it doesn't have the attr, getattr returns None. Otherwise, it has gravity...
            #Of course, we don't even bother checking if there's nothing there at all (theobj will be None)
            if theobj and getattr(theobj,"gravity",None):
                #so we set it.
                theobj.gravity=self.gravadj
    
    def draw(self, window):
        super().draw(window)
        color="gray"
        if not (isinstance(self.gravadj,tuple) and len(self.gravadj)==2 and 
                isinstance(self.gravadj[0],int) and isinstance(self.gravadj[1],int)):
            #If gravadj is invalid, we draw our force lines in red.
            color="red"
        x, y = self.pos
        #For every tile we effect that is empty (nothing's been drawn yet), draw a little thingy there to show
        #our range.
        for yofst in range(1,self.range+1):
            if window.getchar(x, y-yofst)==" ":
                window.putchar("☷",x, y-yofst, fgcolor=color)
        
class SignalSplitter():
    """A trigger that, when tripped by another trigger, triggers more than one target. This is a virtual object; it isn't
    actually a GameObj or Trigger."""
    
    def __init__(self, targets):
        self.targets=targets
        
    def trigger(self):
        for thetarget in self.targets:
            thetarget.trigger()
        
class Missile(GameObj):
    """An object that moves towards its target until it encounters another object (except pressureplates and such).
    Whatever object it collides with is removed from the game, unless it's a wall or the player. Walls are unaffected;
    hitting the player moves him back to where he started the level."""
    
    dir_char_map={(0,1) : "|", (0,-1) : "|", (1,0): "―", (-1,0) : "―", (1,1): "\\", (-1,-1) : "\\", (-1,1) : "/", (1,-1): "/", (0,0) : ""}
    
    def __init__(self,startpos,target):
        super().__init__("|",startpos)
        self.target=target
    
    def update(self,gamestate):
        #tx and ty stand for Target x & y
        #The player can set our target to None, which obviously doesn't have pos and raises AttributeError
        #If this happens, we just destroy ourself (the idea being that the missile, having no target, has crashed
        #into the ground and detonated harmlessly).
        try:
            tx, ty = self.target.int_pos()
        except AttributeError:
            gamestate.current_level.remove_obj(self)
            return
        x, y = self.pos
        
        #Get a direction vector leading to our target (directionx, directiony)
        dx = tx - x
        dy = ty - y
        
        #And clamp it to -1,0,1
        if abs(dx) > 1:
            dx=dx//abs(dx)
        if abs(dy) > 1:
            dy=dy//abs(dy)
            
        #Do our cool face-the-direction-you're-moving effect
        self.graphic=self.dir_char_map[(dx, dy)]
        
        newpos=(x+dx,y+dy)
        
        #If we're about to hit something...
        hit=movement_check(newpos,gamestate.current_level)
        if hit:
            if isinstance(hit, Wall):
                #We don't do anything to walls.
                pass
            elif isinstance(hit, Player):
                #if we hit the player, call its "kill" function
                hit.kill(self, gamestate)
            else:
                #For everything else, there's Mas- err, I mean, destroy it.
                gamestate.current_level.remove_obj(hit)
            
            #In any case, that's the end of this missile.
            gamestate.current_level.remove_obj(self)
            return
        
        
        #And if we haven't hit anything, we can actually move.
        self.pos=newpos
        
class MissileLauncher(GameObj):
    """An object that spawns missiles directed at its target at rate frames per missile (default of 30; game runs at 15 fps).
    If addsafe is True, they will be added to the safe_objlist instead of the main objlist and be unaccessible from the injection
    console."""
    
    def __init__(self,startpos,target,rate=30,addsafe=False):
        super().__init__("☸",startpos)
        self.target=target
        self.refireclock=self.rate=rate
        self.addsafe=addsafe
    
    def update(self,gamestate):
        #If it's time to launch a missile...
        self.refireclock-=1
        if not self.refireclock:
            self.refireclock=self.rate
        else:
            return
        
        #Create it and add it to the appropriate list
        #We make the missile at our own pos; it won't hit us, because it looks ahead
        missile=Missile(self.pos,self.target)
        
        if self.addsafe:
            gamestate.current_level.safe_objlist.append(missile)
        else:
            gamestate.current_level.objlist.append(missile)
            
class SafePressurePlate(PressurePlate):
    """A derivative of the pressure plate that only accepts gates as targets. Makes sure the gate can't be \"forced\" open;
    the only way to make it stay open is to depress the pressure plate."""
    
    def update(self, gamestate):
        #We do nothing if our target isn't a gate
        if not isinstance(self.target, Gate):
            return
        PressurePlate.update(self, gamestate)
        
        #If we're down, the gate is open. Otherwise, keep is shut.
        self.target.open=self.down
        
    def _trip(self):
        #We null out our _trip function 'cos we're not even using it.
        pass
        
class AttributeEnforcer(GameObj):
    """A gameobj that constantly re-sets certain object's attributes to what they are at the beginning of the level.
    attribdict maps tuples of (obj, "param") to None; it'll be populated with values on the first update()."""
    
    def __init__(self,pos,attribdict):
        super().__init__("☑",pos)
        self.attribdict=attribdict
        self.first=True
        #This is a timer/counter that is set to 30 (game runs at 15fps) whenever the object sees one of its watched
        #parameters has been changed and is decremented every frame. If it is non-zero, we're drawn as a red, 
        #crossed-out checkbox, to help the player work out that this object is responsible for re-setting his parameters.
        self.graphicreset=0
        
    def update(self,gamestate):
        #On our first update(), we should grab all the parameters.
        if self.first:
            for key in self.attribdict:
                obj, param = key
                self.attribdict[key]=getattr(obj,param)
            self.first=False
        else:
            #After that, just keep setting those values
            for key in self.attribdict:
                obj, param = key
                if not self.attribdict[key]==getattr(obj,param):
                    self.graphicreset=30
                    setattr(obj,param,self.attribdict[key])
                    
    def draw(self,window):
        x, y=self.pos
        
        if self.graphicreset:
            self.graphicreset-=1
            window.putchar("☒", int(x), int(y), fgcolor="red")
        else:
            window.putchar("☑", int(x), int(y), fgcolor="green")
            
class Teleporter(GameObj):
    "Anything that enters the same square as this object is teleported to its dest coordinates."
    
    def __init__(self,pos,dest):
        super().__init__("˽",pos)
        self.dest=dest
        
    def update(self,gamestate):
        occupant=gamestate.current_level.get_at(self.pos)
        if occupant:
            occupant.pos=self.dest

class DroneProxy():
    """The object a Drone passes into its think() function instead of itself. This allows the player-written code to, say,
    read the drone's position, but not set it."""
            
            
    def __init__(self,drone):
        if not isinstance(drone,Drone):
            raise TypeError("Tried to construct a DroneProxy around a {0}: not a gameobjs.Drone!".format(drone.__class__))
        self.drone=drone
    
    #The property decorator here just makes the attribute read-only
    @property
    def pos(self):
        "The drone is at this position on the screen. Read-only."
        return self.drone.pos
    
    @property
    def bumped(self):
        "If the drone tried to move into occupied space last turn, this references the object that was occupying that space."
        return self.drone.bumped
    
    @property
    def color(self):
        "The drone's icon will be drawn in this color. Read-write."
        return self.drone.color
    
    @color.setter
    def color(self,value):
        self.drone.color=value
        
    def move(self,vec):
        self.drone.move(vec)

        
        
            
class Drone(GameObj):
    """
    An object that can move around and trip pressureplates and such, controlled by a user-defined think() function. 
    Be careful; it can only move once per tick (think is executed once each tick)!
    
    The think function should accept only one parameter; self. Think() does not have access to all the attributes and methods 
    of its drone. The API it does get is as such:
    
    methods:
    self.move(vec) : Move the drone in the given direction (i.e. move((0,1)) moves it down the screen, (1,0) moves it right)
    
    attributes:
    self.pos : A 2-tuple containing the drone's position.
    self.color : A 3-tuple of RGB values; the drone icon will be this color.
    self.bumped : If the drone bumped into something last tick (tried to move into the space it occupied), this references that object.
    
    think() can also put new attributes on self if it wants to have persistent state.
    
    Anything that normally sends text to stdout will print to the injection console.
    Drones should always be deployed with AttributeEnforcer-s that stop the player from modifying functions other than think(),
    or a DroneUploadSystem.
    """
    
    def __init__(self,pos,think=null_func):
        super().__init__("♙",pos)
        self.move_vec=(0,0)
        self.think=think
        self.color=(255,255,255)
        self.proxy=DroneProxy(self)
        self.bumped=None
    
    def update(self,gamestate):
        #Note that since think is always monkey-patched, always defined outside the Drone class,
        #it does not get the self parameter. 
        self.think(self.proxy)
        
        if self.move_vec[0] or self.move_vec[1]:
            dx, dy = self.move_vec
            self.move_vec=(0,0)
            newpos=(self.pos+dx,self.pos+dy)
            
            self.bumped=movement_check(newpos,gamestate.current_level)
            if self.bumped:
                return
            else:
                self.pos=newpos
                return True
        
    def draw(self, window):
        window.putchar(" ", fgcolor=self.color)
        super().draw(window)
        
    def move(self,vec):
        """Try to move along a given vector tuple. Returns nothing. Note that the movement only takes effect the frame after move() is
        called, and only the last move() command actually does anything."""
        
        #We can only move once per tick; update() will reset this each tick
        if self.moved:
            return False
        
        #d is for delta
        dx, dy = vec
        
        #clamping again
        if abs(dx) > 1:
            dx=dx//abs(dx)
        if abs(dy) > 1:
            dy=dy//abs(dy)
            
        #We can't actually do the things needed to move ourself without map data, so we wait 'till next update.
        self.move_vec=(dx, dy)
        
    def upload(self,func):
        self.think=func
    upload.__doc__="""Set this drone's think() to the given function. This is added to the injection 
        console namespace in leiu of the drone itself. The following is a description of the drone 
        API, which is also described in the readme:
        
        drone.think is passed one parameter, self, which is a proxy to the drone object.
        
        self.move(vec) : Move the drone in the given direction 
        (i.e. move((0,1)) moves it down the screen, (1,0) moves it right)
        
        self.pos : A 2-tuple containing the drone's position.
        self.color : A 3-tuple of RGB values; the drone icon will be this color.
        self.bumped : If the drone bumped into something last tick 
        (tried to move into the space it occupied), this references that object.
        """.replace("    ","")
        
#DEPRECATED
#Never worked in the first place, supplanted by drone.upload
class DroneUploadSystem(GameObj):
    """This object allows the player to set a drone's think function without actually having direct access to the drone.
    It also allows the player access to the drone obj's docstr, which contains information about the drone API.
    This is a virtual game object; it doesn't actually show up in the world.
    DEPRECATED: This never worked in the first place"""
    
    dronedoc=Drone.__doc__
    
    def __init__(self, _drone):
        super().__init__("",(-1,-1))
        #This bit of wizardry ensures that only functions named "upload" can read the attr.
        #If the player really wanted, he could spoof this, but at that point he deserves the damn drone object
        #The function here has access to _drone because closures ugh so tired
        def dronegetter():
            callername=inspect.stack()[1][3]
            if callername=="upload" or callername=="__getstate__":
                return _drone
            else:
                print("What part of \"without actually having direct access to the drone\" didn't you understand?")
                return None
        self.drone=dronegetter
        del dronegetter
        
    
    def __getstate__(self):
        #We have to return our dict sans all the drone access fuckery because Pickle doesn't like it
        #when functions are defined outside of the top level of a module
        mydict=self.__dict__
        del mydict["getter"]
        mydict["_drone"]=self.drone
        return
    
    def __setstate__(self,state):
        drone=state["_drone"]
        self.__init__(drone)
        del state["_drone"]
        self.__dict__=state 
    
    def upload(self,func):
        self.drone.think=func
        
    @property
    def drone(self):
        #The property definition in the class definition does nothing
        pass
    
    @drone.setter
    def drone(self,func):
        "This actually sets the getter function, not the value of drone. Yeah."
        callername=inspect.stack()[1][3]
        if callername=="__init__":
            self.getter=func
        
class CombatAndroid(GameObj):
    "An object that hunts down enemy objects and shoots them with deadly lasers."
    
    def __init__(self,pos,targets=[]):
        super().__init__("©",pos)
        self.targets=targets
        self.target=None
        
    def is_enemy(self,obj):
        "Determine whether or not the given object is an enemy unit and should be destroyed. If it is, returns true."
        
        if isinstance(obj,Player):
            return True
        elif obj in self.targets:
            return True
        else:
            return False
        
    def update(self,gamestate):
        #If our target is None, we need a new one.
        if not self.target:
            #Build an iterator (ptgts) containing all active gameobjs that are enemies
            #(self.is_enemy() returns true on them)
            ptgts1=(obj for obj in gamestate.current_level.objlist if self.is_enemy(obj))
            ptgts2=(obj for obj in gamestate.current_level.safe_objlist if self.is_enemy(obj))
            if self.is_enemy(gamestate.current_level.plyobj):
                ptgts=itertools.chain(ptgts1,ptgts2,(gamestate.current_level.plyobj,))
            else:
                ptgts=itertools.chain(ptgts1,ptgts2)
            
            current_target=None
            current_dist=float("+inf")
            #find closest object in ptgts
            for obj in ptgts:
                objdist=distance(obj, self)
                if objdist < current_dist:
                    current_target=obj
                    current_dist=objdist
            self.target=current_target
            
        #Now we try shooting at our target...
        if raycast_to(self.int_pos(), self.target.int_pos(), gamestate.current_level) == self.target:
            print("Can see target!")
        #print(path(self.int_pos(),self.target.int_pos(), gamestate.current_level, gamestate.mainwin))
                    
class LaserTurret(GameObj):
            "An object that fires deadly lasers at the nearest enemy target."
            
            def __init__(self,pos,targets=[]):
                super().__init__('⌀',pos)
                self.target=None
                self.targets=targets
                
            def is_enemy(self,obj):
                "Determine whether or not the given object is an enemy unit and should be destroyed. If it is, returns true."
                
                if isinstance(obj,Player):
                    return True
                elif obj in self.targets:
                    return True
                else:
                    return False
                        
            def update(self,gamestate):
                #If our target is None, we need a new one.
                if not self.target:
                    #Build an iterator (ptgts) containing all active gameobjs that are enemies
                    #(self.is_enemy() returns true on them)
                    ptgts1=(obj for obj in gamestate.current_level.objlist if self.is_enemy(obj))
                    ptgts2=(obj for obj in gamestate.current_level.safe_objlist if self.is_enemy(obj))
                    if self.is_enemy(gamestate.current_level.plyobj):
                        ptgts=itertools.chain(ptgts1,ptgts2,(gamestate.current_level.plyobj,))
                    else:
                        ptgts=itertools.chain(ptgts1,ptgts2)
                    
                    current_target=None
                    current_dist=float("+inf")
                    #find closest object in ptgts
                    for obj in ptgts:
                        objdist=distance(obj, self)
                        if objdist < current_dist:
                            current_target=obj
                            current_dist=objdist
                    self.target=current_target
                    print("Target is",self.target)
                    
                if raycast_to(self.int_pos(),self.target.int_pos(),gamestate.current_level) == self.target:
                    hit = laser(self.int_pos(), self.target.int_pos(), gamestate)
                    #We only harm the player; all other objects are unaffected
                    if hit and isinstance(hit,Player):
                        hit.kill(self,gamestate)
                else:
                    self.target=None
                    
class LaserBeam(GameObj):
    "A purely graphical object that's responsible for drawing awesome laz0r beams."
    
    def __init__(self,src,dest,color,lifetime=5):
        super().__init__("",(-1,-1))
        self.src=src
        self.dest=dest
        self.color=color
        self.lifetime=lifetime
        
    def update(self,gamestate):
        if self.lifetime:
            self.lifetime-=1
        else:
            gamestate.current_level.remove_obj(self)
            
    def draw(self,window):
        surface=window.surface
        pygame.draw.line(surface,self.color,self.src,self.dest)


class Lvl1Keypad(Trigger):
    
    def __init__(self,startingpos,target):
        super().__init__('▫',startingpos,target)
        self.open=False
        #brute this muthafucka
        self.code=str(random.randint(10**128,10**129-1))
    
    def interact(self,player,gamestate):
        player.pos=player.oldpos
        win=gamestate.mainwin
        win.fill()
        win.pygprint("Please enter your 128-digit authorization code:")
        win.pygprint(">")
        win.update()
        code=win.input()
        if code == self.code:
            self._trip()
            win.pygprint("Authorization code accepted!")
        else:
            win.pygprint("Invalid authorization code.")
        gamestate.mainwin.update()
        pygcurse.waitforkeypress()
        
#Objects of these classes are considered passable; object such as boulders and missiles can move through them.
no_collide_objs=(PressurePlate,GravPad)
        